"""Result and Option interfaces."""

import typing as t

# pylint: disable=invalid-name

T = t.TypeVar("T")
E = t.TypeVar("E")
U = t.TypeVar("U")
F = t.TypeVar("F")

ExcType = t.TypeVar("ExcType", bound=Exception)


class _Result(t.Generic[T, E]):
    """Standard wrapper for results."""

    __slots__ = ()

    def __init__(self, result: t.Union[T, E]) -> None:
        """Results may not be instantiated directly."""
        raise NotImplementedError(
            "Results are only for type hinting and may not be instantiated "
            "directly. Please use Ok() or Err() instead."
        )

    # ------------------------------------------------------------------
    # Constructors
    # ------------------------------------------------------------------

    # See https://github.com/python/mypy/issues/3737 for issue with
    # specifying a default type. However, type hinting of uses of this
    # method should still work just fine.
    @staticmethod
    def of(
        fn: t.Callable[..., T],
        *args: t.Any,
        catch: t.Type[ExcType] = Exception,  # type: ignore
        **kwargs: t.Any
    ) -> "_Result[T, ExcType]":
        """Call `fn` and wrap its result in an `Ok()`.

        If an exception is intercepted, return `Err(exception)`. By
        default, any `Exception` will be intercepted. If you specify
        `exc_type`, only that exception will be intercepted.
        """
        raise NotImplementedError

    @staticmethod
    def err_if(predicate: t.Callable[[T], bool], value: T) -> "_Result[T, T]":
        """Return Err(val) if predicate(val) is True, otherwise Ok(val)."""
        raise NotImplementedError

    @staticmethod
    def ok_if(predicate: t.Callable[[T], bool], value: T) -> "_Result[T, T]":
        """Return Ok(val) if predicate(val) is True, otherwise Err(val)."""
        raise NotImplementedError

    # ------------------------------------------------------------------
    # Methods
    # ------------------------------------------------------------------

    def and_(self, res: "_Result[U, E]") -> "_Result[U, E]":
        """Return `res` if self is `Ok`, otherwise return `self`."""
        raise NotImplementedError

    def or_(self, res: "_Result[T, F]") -> "_Result[T, F]":
        """Return `res` if self is `Err`, otherwise `self`."""
        raise NotImplementedError

    def and_then(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error. Alias of `flatmap`.

        This can be used to chain functions that return results.
        """
        raise NotImplementedError

    def flatmap(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error. Alias of `and_then`

        This can be used to chain functions that return results.
        """
        raise NotImplementedError

    def or_else(self, fn: t.Callable[[E], "_Result[T, F]"]) -> "_Result[T, F]":
        """Return `self` if `Ok`, or call `fn` with `self` if `Err`."""
        raise NotImplementedError

    def err(self) -> "_Option[E]":
        """Return Err value if result is Err."""
        raise NotImplementedError

    def ok(self) -> "_Option[T]":
        """Return OK value if result is Ok."""
        raise NotImplementedError

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Return `Ok` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        raise NotImplementedError

    def expect_err(
        self, msg: str, exc_cls: t.Type[Exception] = RuntimeError
    ) -> E:
        """Return `Err` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        raise NotImplementedError

    def is_err(self) -> bool:
        """Returl whether the result is an Err."""
        raise NotImplementedError

    def is_ok(self) -> bool:
        """Return whether the result is OK."""
        raise NotImplementedError

    def iter(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        raise NotImplementedError

    def map(self, fn: t.Callable[[T], U]) -> "_Result[U, E]":
        """Map a function onto an okay result, or ignore an error."""
        raise NotImplementedError

    def map_err(self, fn: t.Callable[[E], F]) -> "_Result[T, F]":
        """Map a function onto an error, or ignore a success."""
        raise NotImplementedError

    def unwrap(self) -> T:
        """Return an Ok result, or throw an error if an Err."""
        raise NotImplementedError

    def unwrap_err(self) -> E:
        """Return an Ok result, or throw an error if an Err."""
        raise NotImplementedError

    def unwrap_or(self, alternative: U) -> t.Union[T, U]:
        """Return the `Ok` value, or `alternative` if `self` is `Err`."""
        raise NotImplementedError

    def unwrap_or_else(self, fn: t.Callable[[E], U]) -> t.Union[T, U]:
        """Return the `Ok` value, or the return from `fn`."""
        raise NotImplementedError

    def __iter__(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        raise NotImplementedError

    def __eq__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        raise NotImplementedError

    def __ne__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Return string value of result."""
        raise NotImplementedError

    def __repr__(self) -> str:
        """Return repr for result."""
        raise NotImplementedError


class _Option(t.Generic[T]):
    """A value that may be `Some` or `Nothing`."""

    __slots__ = ()

    def __init__(self, value: t.Optional[T]) -> None:
        """Options may not be instantiated directly."""
        raise NotImplementedError(
            "Options may not be instantiated directly. Use Some() or "
            "Nothing() instead."
        )

    # ------------------------------------------------------------------
    # Constructors
    # ------------------------------------------------------------------

    @staticmethod
    def of(value: t.Optional[T]) -> "_Option[T]":
        """Construct an _Option[T] from an Optional[T]."""
        raise NotImplementedError

    @staticmethod
    def nothing_if(predicate: t.Callable[[T], bool], value: T) -> "_Option[T]":
        """Return Nothing() if predicate(val) is True, else Some(val)."""
        raise NotImplementedError

    @staticmethod
    def some_if(predicate: t.Callable[[T], bool], value: T) -> "_Option[T]":
        """Return Some(val) if predicate(val) is True, else Nothing()."""
        raise NotImplementedError

    # ------------------------------------------------------------------
    # Methods
    # ------------------------------------------------------------------

    def and_(self, alternative: "_Option[U]") -> "_Option[U]":
        """Return `Nothing` if `self` is `Nothing`, or the `alternative`."""
        raise NotImplementedError

    def or_(self, alternative: "_Option[T]") -> "_Option[T]":
        """Return option if it is `Some`, or the `alternative`."""
        raise NotImplementedError

    def xor(self, alternative: "_Option[T]") -> "_Option[T]":
        """Return Some IFF exactly one of `self`, `alternative` is `Some`."""
        raise NotImplementedError

    def and_then(self, fn: t.Callable[[T], "_Option[U]"]) -> "_Option[U]":
        """Return `Nothing`, or call `fn` with the `Some` value."""
        raise NotImplementedError

    def flatmap(self, fn: t.Callable[[T], "_Option[U]"]) -> "_Option[U]":
        """Return `Nothing`, or call `fn` with the `Some` value."""
        raise NotImplementedError

    def or_else(self, fn: t.Callable[[], "_Option[T]"]) -> "_Option[T]":
        """Return option if it is `Some`, or calculate an alternative."""
        raise NotImplementedError

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Unwrap and yield a `Some`, or throw an exception if `Nothing`.

        The exception class may be specified with the `exc_cls` keyword
        argument.
        """
        raise NotImplementedError

    def filter(self, predicate: t.Callable[[T], bool]) -> "_Option[T]":
        """Return `Nothing`, or an option determined by the predicate.

        If `self` is `Some`, call `predicate` with the wrapped value and
        return:

        * `self` (`Some(t)` where `t` is the wrapped value) if the predicate
          is `True`
        * `Nothing` if the predicate is `False`
        """
        raise NotImplementedError

    def is_nothing(self) -> bool:
        """Return whether the option is `Nothing`."""
        raise NotImplementedError

    def is_some(self) -> bool:
        """Return whether the option is a `Some` value."""
        raise NotImplementedError

    def iter(self) -> t.Iterator[T]:
        """Return an iterator over the possibly contained value."""
        raise NotImplementedError

    def map(self, fn: t.Callable[[T], U]) -> "_Option[U]":
        """Apply `fn` to the contained value if any."""
        raise NotImplementedError

    def map_or(self, default: U, fn: t.Callable[[T], U]) -> U:
        """Apply `fn` to contained value, or return the default."""
        raise NotImplementedError

    def map_or_else(
        self, default: t.Callable[[], U], fn: t.Callable[[T], U]
    ) -> U:
        """Apply `fn` to contained value, or compute a default."""
        raise NotImplementedError

    def ok_or(self, err: E) -> _Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err)`.
        """
        raise NotImplementedError

    def ok_or_else(self, err_fn: t.Callable[[], E]) -> _Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err_fn())`.
        """
        raise NotImplementedError

    def unwrap(self) -> T:
        """Return `Some` value, or raise an error."""
        raise NotImplementedError

    def unwrap_or(self, default: U) -> t.Union[T, U]:
        """Return the contained value or `default`."""
        raise NotImplementedError

    def unwrap_or_else(self, fn: t.Callable[[], U]) -> t.Union[T, U]:
        """Return the contained value or calculate a default."""
        raise NotImplementedError

    def __iter__(self) -> t.Iterator[T]:
        """Iterate over the contained value if present."""
        raise NotImplementedError

    def __eq__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        raise NotImplementedError

    def __ne__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Return a string representation of the Option."""
        raise NotImplementedError

    def __repr__(self) -> str:
        """Return a string representation of the Option."""
        raise NotImplementedError



"""Implementations of Ok, Err, Some, and None."""

import typing as t

from ._interface import _Option, _Result


T = t.TypeVar("T")
E = t.TypeVar("E")
U = t.TypeVar("U")
F = t.TypeVar("F")

ExcType = t.TypeVar("ExcType", bound=Exception)

WrappedFunc = t.Callable[..., t.Any]
WrappedFn = t.TypeVar("WrappedFn", bound=WrappedFunc)

Args = t.TypeVar("Args")
Kwargs = t.TypeVar("Kwargs")

# pylint: disable=super-init-not-called


# pylint: disable=abstract-method
class Result(_Result[T, E]):
    """Base implementation for Result types."""

    __slots__ = ()

    @staticmethod
    def of(
        fn: t.Callable[..., T],
        *args: t.Any,
        catch: t.Type[ExcType] = Exception,  # type: ignore
        **kwargs: t.Any,
    ) -> "Result[T, ExcType]":
        """Call `fn` and wrap its result in an `Ok()`.

        If an exception is intercepted, return `Err(exception)`. By
        default, any `Exception` will be intercepted. If you specify
        `exc_type`, only that exception will be intercepted.
        """
        try:
            return Ok(fn(*args, **kwargs))
        except catch as exc:
            return Err(exc)

    @staticmethod
    def err_if(predicate: t.Callable[[T], bool], value: T) -> "Result[T, T]":
        """Return Err(val) if predicate(val) is True, otherwise Ok(val)."""
        if predicate(value):
            return Err(value)
        return Ok(value)

    @staticmethod
    def ok_if(predicate: t.Callable[[T], bool], value: T) -> "Result[T, T]":
        """Return Ok(val) if predicate(val) is True, otherwise Err(val)."""
        if predicate(value):
            return Ok(value)
        return Err(value)


class Option(_Option[T]):
    """Base implementation for Option types."""

    __slots__ = ()

    @staticmethod
    def of(value: t.Optional[T]) -> "Option[T]":
        """Construct an Option[T] from an Optional[T].

        If the value is None, Nothing() is returned. If the value is
        not None, Some(value) is returned.
        """
        if value is None:
            return Nothing()
        return Some(value)

    @staticmethod
    def nothing_if(predicate: t.Callable[[T], bool], value: T) -> "Option[T]":
        """Return Nothing() if predicate(val) is True, else Some(val)."""
        if predicate(value):
            return Nothing()
        return Some(value)

    @staticmethod
    def some_if(predicate: t.Callable[[T], bool], value: T) -> "Option[T]":
        """Return Some(val) if predicate(val) is True, else Nothing()."""
        if predicate(value):
            return Some(value)
        return Nothing()


# pylint: enable=abstract-method


class Ok(Result[T, E]):
    """Standard wrapper for results."""

    __slots__ = ("_value",)

    def __init__(self, result: T) -> None:
        """Wrap a result."""
        self._value: T = result

    def and_(self, res: "_Result[U, E]") -> "_Result[U, E]":
        """Return `res` if the result is `Ok`, otherwise return `self`."""
        return res

    def or_(self, res: "_Result[T, F]") -> "_Result[T, F]":
        """Return `res` if the result is `Err`, otherwise `self`."""
        return t.cast(Result[T, F], self)

    def and_then(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error.

        This can be used to chain functions that return results.
        """
        return fn(self._value)

    def flatmap(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error.

        This can be used to chain functions that return results.
        """
        return self.and_then(fn)

    def or_else(self, fn: t.Callable[[E], "_Result[T, F]"]) -> "_Result[T, F]":
        """Return `self` if `Ok`, or call `fn` with `self` if `Err`."""
        return t.cast(Result[T, F], self)

    def err(self) -> _Option[E]:
        """Return Err value if result is Err."""
        return Nothing()

    def ok(self) -> _Option[T]:
        """Return OK value if result is Ok."""
        return Some(self._value)

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Return `Ok` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        return self._value

    def expect_err(
        self, msg: str, exc_cls: t.Type[Exception] = RuntimeError
    ) -> E:
        """Return `Err` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        raise exc_cls(msg)

    def is_err(self) -> bool:
        """Returl whether the result is an Err."""
        return False

    def is_ok(self) -> bool:
        """Return whether the result is OK."""
        return True

    def iter(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        return iter(self)

    def map(self, fn: t.Callable[[T], U]) -> "_Result[U, E]":
        """Map a function onto an okay result, or ignore an error."""
        return Ok(fn(self._value))

    def map_err(self, fn: t.Callable[[E], F]) -> "_Result[T, F]":
        """Map a function onto an error, or ignore a success."""
        return t.cast(Result[T, F], self)

    def unwrap(self) -> T:
        """Return an Ok result, or throw an error if an Err."""
        return self._value

    def unwrap_err(self) -> E:
        """Return an Ok result, or throw an error if an Err."""
        raise RuntimeError(f"Tried to unwrap_err {self}!")

    def unwrap_or(self, alternative: U) -> t.Union[T, U]:
        """Return the `Ok` value, or `alternative` if `self` is `Err`."""
        return self._value

    def unwrap_or_else(self, fn: t.Callable[[E], U]) -> t.Union[T, U]:
        """Return the `Ok` value, or the return from `fn`."""
        return self._value

    def __iter__(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        yield self._value

    def __eq__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        if not isinstance(other, Result):
            return False
        if not other.is_ok():
            return False
        eq: bool = self._value == other.unwrap()
        return eq

    def __ne__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        return not self == other

    def __str__(self) -> str:
        """Return string value of result."""
        return f"{self.__class__.__name__}({repr(self._value)})"

    def __repr__(self) -> str:
        """Return repr for result."""
        return self.__str__()


class Err(Result[T, E]):
    """Standard wrapper for results."""

    __slots__ = ("_value",)

    def __init__(self, result: E) -> None:
        """Wrap a result."""
        self._value = result

    def and_(self, res: "_Result[U, E]") -> "_Result[U, E]":
        """Return `res` if the result is `Ok`, otherwise return `self`."""
        return t.cast(Result[U, E], self)

    def or_(self, res: "_Result[T, F]") -> "_Result[T, F]":
        """Return `res` if the result is `Err`, otherwise `self`."""
        return res

    def and_then(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error.

        This can be used to chain functions that return results.
        """
        return t.cast(Result[U, E], self)

    def flatmap(self, fn: t.Callable[[T], "_Result[U, E]"]) -> "_Result[U, E]":
        """Call `fn` if Ok, or ignore an error.

        This can be used to chain functions that return results.
        """
        return self.and_then(fn)

    def or_else(self, fn: t.Callable[[E], "_Result[T, F]"]) -> "_Result[T, F]":
        """Return `self` if `Ok`, or call `fn` with `self` if `Err`."""
        return fn(self._value)

    def err(self) -> _Option[E]:
        """Return Err value if result is Err."""
        return Some(self._value)

    def ok(self) -> _Option[T]:
        """Return OK value if result is Ok."""
        return Nothing()

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Return `Ok` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        raise exc_cls(msg)

    def expect_err(
        self, msg: str, exc_cls: t.Type[Exception] = RuntimeError
    ) -> E:
        """Return `Err` value or raise an error with the specified message.

        The raised exception class may be specified with the `exc_cls`
        keyword argument.
        """
        return self._value

    def is_err(self) -> bool:
        """Returl whether the result is an Err."""
        return True

    def is_ok(self) -> bool:
        """Return whether the result is OK."""
        return False

    def iter(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        return iter(self)

    def map(self, fn: t.Callable[[T], U]) -> "_Result[U, E]":
        """Map a function onto an okay result, or ignore an error."""
        return t.cast(Result[U, E], self)

    def map_err(self, fn: t.Callable[[E], F]) -> "_Result[T, F]":
        """Map a function onto an error, or ignore a success."""
        return Err(fn(self._value))

    def unwrap(self) -> T:
        """Return an Ok result, or throw an error if an Err."""
        raise RuntimeError(f"Tried to unwrap {self}!")

    def unwrap_err(self) -> E:
        """Return an Ok result, or throw an error if an Err."""
        return self._value

    def unwrap_or(self, alternative: U) -> t.Union[T, U]:
        """Return the `Ok` value, or `alternative` if `self` is `Err`."""
        return alternative

    def unwrap_or_else(self, fn: t.Callable[[E], U]) -> t.Union[T, U]:
        """Return the `Ok` value, or the return from `fn`."""
        return fn(self._value)

    def __iter__(self) -> t.Iterator[T]:
        """Return a one-item iterator whose sole member is the result if `Ok`.

        If the result is `Err`, the iterator will contain no items.
        """
        _: t.Tuple[T, ...] = ()
        yield from _

    def __eq__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        if not isinstance(other, Result):
            return False
        if not other.is_err():
            return False
        eq: bool = self._value == other.unwrap_err()
        return eq

    def __ne__(self, other: t.Any) -> bool:
        """Compare two results. They are equal if their values are equal."""
        return not self == other

    def __str__(self) -> str:
        """Return string value of result."""
        return f"{self.__class__.__name__}({repr(self._value)})"

    def __repr__(self) -> str:
        """Return repr for result."""
        return self.__str__()


class Some(Option[T]):
    """A value that may be `Some` or `Nothing`."""

    __slots__ = ("_value",)

    def __init__(self, value: T) -> None:
        """Wrap value in a `Some()`."""
        # not sure why pylint things _value is not in __slots__
        self._value = value  # pylint: disable=assigning-non-slot

    def and_(self, alternative: _Option[U]) -> _Option[U]:
        """Return `Nothing` if `self` is `Nothing`, or the `alternative`."""
        return alternative

    def or_(self, alternative: _Option[T]) -> _Option[T]:
        """Return option if it is `Some`, or the `alternative`."""
        return self

    def xor(self, alternative: _Option[T]) -> _Option[T]:
        """Return Some IFF exactly one of `self`, `alternative` is `Some`."""
        return (
            t.cast(_Option[T], self) if alternative.is_nothing() else Nothing()
        )

    def and_then(self, fn: t.Callable[[T], _Option[U]]) -> _Option[U]:
        """Return `Nothing`, or call `fn` with the `Some` value."""
        return fn(self._value)

    def flatmap(self, fn: t.Callable[[T], "_Option[U]"]) -> "_Option[U]":
        """Return `Nothing`, or call `fn` with the `Some` value."""
        return self.and_then(fn)

    def or_else(self, fn: t.Callable[[], _Option[T]]) -> _Option[T]:
        """Return option if it is `Some`, or calculate an alternative."""
        return self

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Unwrap and yield a `Some`, or throw an exception if `Nothing`.

        The exception class may be specified with the `exc_cls` keyword
        argument.
        """
        return self._value

    def filter(self, predicate: t.Callable[[T], bool]) -> _Option[T]:
        """Return `Nothing`, or an option determined by the predicate.

        If `self` is `Some`, call `predicate` with the wrapped value and
        return:

        * `self` (`Some(t)` where `t` is the wrapped value) if the predicate
          is `True`
        * `Nothing` if the predicate is `False`
        """
        if predicate(self._value):
            return self
        return Nothing()

    def is_nothing(self) -> bool:
        """Return whether the option is `Nothing`."""
        return False

    def is_some(self) -> bool:
        """Return whether the option is a `Some` value."""
        return True

    def iter(self) -> t.Iterator[T]:
        """Return an iterator over the possibly contained value."""
        return iter(self)

    def map(self, fn: t.Callable[[T], U]) -> _Option[U]:
        """Apply `fn` to the contained value if any."""
        return Some(fn(self._value))

    def map_or(self, default: U, fn: t.Callable[[T], U]) -> U:
        """Apply `fn` to contained value, or return the default."""
        return fn(self._value)

    def map_or_else(
        self, default: t.Callable[[], U], fn: t.Callable[[T], U]
    ) -> U:
        """Apply `fn` to contained value, or compute a default."""
        return fn(self._value)

    def ok_or(self, err: E) -> Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err)`.
        """
        res: Result[T, E] = Ok(t.cast(T, self._value))
        return res

    def ok_or_else(self, err_fn: t.Callable[[], E]) -> Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err_fn())`.
        """
        res: Result[T, E] = Ok(t.cast(T, self._value))
        return res

    def unwrap(self) -> T:
        """Return `Some` value, or raise an error."""
        return self._value

    def unwrap_or(self, default: U) -> t.Union[T, U]:
        """Return the contained value or `default`."""
        return self._value

    def unwrap_or_else(self, fn: t.Callable[[], U]) -> t.Union[T, U]:
        """Return the contained value or calculate a default."""
        return self._value

    def __iter__(self) -> t.Iterator[T]:
        """Iterate over the contained value if present."""
        yield self._value

    def __eq__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        if not isinstance(other, Some):
            return False
        eq: bool = self._value == other.unwrap()
        return eq

    def __ne__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        return not self == other

    def __str__(self) -> str:
        """Represent the Some() as a string."""
        return f"Some({repr(self._value)})"

    def __repr__(self) -> str:
        """Return a string representation of the Some()."""
        return self.__str__()


class Nothing(Option[T]):
    """A value that may be `Some` or `Nothing`."""

    __slots__ = ("_value",)

    _instance = None

    def __init__(self, _: None = None) -> None:
        """Create a Nothing()."""
        if self._instance is None:
            # The singleton is being instantiated the first time
            # not sure why pylint things _value is not in __slots__
            self._value = None  # pylint: disable=assigning-non-slot

    def __new__(cls, _: None = None) -> "Nothing[T]":
        """Ensure we are a singleton."""
        if cls._instance is None:
            # Create the instance
            inst = super().__new__(cls)
            # And instantiate it
            cls.__init__(inst)
            # Then assign it to the class' _instance var, so no other
            # instances can be created
            cls._instance = inst
        return t.cast("Nothing[T]", cls._instance)

    def and_(self, alternative: _Option[U]) -> _Option[U]:
        """Return `Nothing` if `self` is `Nothing`, or the `alternative`."""
        return t.cast(_Option[U], self)

    def or_(self, alternative: _Option[T]) -> _Option[T]:
        """Return option if it is `Some`, or the `alternative`."""
        return alternative

    def xor(self, alternative: _Option[T]) -> _Option[T]:
        """Return Some IFF exactly one of `self`, `alternative` is `Some`."""
        return alternative if alternative.is_some() else self

    def and_then(self, fn: t.Callable[[T], _Option[U]]) -> _Option[U]:
        """Return `Nothing`, or call `fn` with the `Some` value."""
        return t.cast(_Option[U], self)

    def flatmap(self, fn: t.Callable[[T], "_Option[U]"]) -> "_Option[U]":
        """Return `Nothing`, or call `fn` with the `Some` value."""
        return self.and_then(fn)

    def or_else(self, fn: t.Callable[[], _Option[T]]) -> _Option[T]:
        """Return option if it is `Some`, or calculate an alternative."""
        return fn()

    def expect(self, msg: str, exc_cls: t.Type[Exception] = RuntimeError) -> T:
        """Unwrap and yield a `Some`, or throw an exception if `Nothing`.

        The exception class may be specified with the `exc_cls` keyword
        argument.
        """
        raise exc_cls(msg)

    def filter(self, predicate: t.Callable[[T], bool]) -> _Option[T]:
        """Return `Nothing`, or an option determined by the predicate.

        If `self` is `Some`, call `predicate` with the wrapped value and
        return:

        * `self` (`Some(t)` where `t` is the wrapped value) if the predicate
          is `True`
        * `Nothing` if the predicate is `False`
        """
        return self

    def is_nothing(self) -> bool:
        """Return whether the option is `Nothing`."""
        return True

    def is_some(self) -> bool:
        """Return whether the option is a `Some` value."""
        return False

    def iter(self) -> t.Iterator[T]:
        """Return an iterator over the possibly contained value."""
        return iter(self)

    def map(self, fn: t.Callable[[T], U]) -> _Option[U]:
        """Apply `fn` to the contained value if any."""
        return t.cast(_Option[U], self)

    def map_or(self, default: U, fn: t.Callable[[T], U]) -> U:
        """Apply `fn` to contained value, or return the default."""
        return default

    def map_or_else(
        self, default: t.Callable[[], U], fn: t.Callable[[T], U]
    ) -> U:
        """Apply `fn` to contained value, or compute a default."""
        return default()

    def ok_or(self, err: E) -> Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err)`.
        """
        res: Result[T, E] = Err(err)
        return res

    def ok_or_else(self, err_fn: t.Callable[[], E]) -> Result[T, E]:
        """Transform an option into a `Result`.

        Maps `Some(v)` to `Ok(v)` or `None` to `Err(err_fn())`.
        """
        res: Result[T, E] = Err(err_fn())
        return res

    def unwrap(self) -> T:
        """Return `Some` value, or raise an error."""
        raise RuntimeError("Tried to unwrap Nothing")

    def unwrap_or(self, default: U) -> t.Union[T, U]:
        """Return the contained value or `default`."""
        return default

    def unwrap_or_else(self, fn: t.Callable[[], U]) -> t.Union[T, U]:
        """Return the contained value or calculate a default."""
        return fn()

    def __iter__(self) -> t.Iterator[T]:
        """Iterate over the contained value if present."""
        _: t.Tuple[T, ...] = ()
        yield from _

    def __eq__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        if not isinstance(other, Nothing):
            return False
        return True

    def __ne__(self, other: t.Any) -> bool:
        """Options are equal if their values are equal."""
        return not self == other

    def __str__(self) -> str:
        """Return a string representation of Nothing()."""
        return f"Nothing()"

    def __repr__(self) -> str:
        """Return a string representation of Nothing()."""
        return self.__str__()



"""Typesafe python versions of Rust-inspired result types."""

__all__ = ("Option", "Result", "Ok", "Err", "Some", "Nothing")
__version__ = "1.0.0"
__version_info__ = tuple(map(int, __version__.split(".")))


from ._impl import Option, Result, Ok, Err, Some, Nothing



# -*- coding: utf-8 -*-
"""Setup file for the skelethon."""

import typing as t
from os.path import dirname, exists, join, realpath
from setuptools import setup, find_packages

cwd = dirname(realpath(__file__))

########################################################################
# Contact Information
########################################################################

URL = "https://www.github.com/mplanchard/safetywrap"
AUTHOR = "Matthew Planchard"
EMAIL = "msplanchard@gmail.com"


########################################################################
# Package Description
########################################################################

NAME = "safetywrap"
SHORT_DESC = "Rust-inspired typesafe result types"

with open(join(cwd, "README.md")) as readme:
    LONG_DESC = readme.read()
LONG_DESC_CONTENT_TYPE = "text/markdown"

KEYWORDS = [
    "python",
    "rust",
    "result",
    "option",
    "typed",
    "types",
    "typesafe",
    "monad",
    "wrapper",
    "safety",
]
CLASSIFIERS = [
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers for all
    # available setup classifiers
    # "Development Status :: 1 - Planning",
    # 'Development Status :: 2 - Pre-Alpha',
    # "Development Status :: 3 - Alpha",
    "Development Status :: 4 - Beta",
    # 'Development Status :: 5 - Production/Stable',
    # 'Development Status :: 6 - Mature',
    # 'Framework :: AsyncIO',
    # 'Framework :: Flask',
    # 'Framework :: Sphinx',
    # 'Environment :: Web Environment',
    "Intended Audience :: Developers",
    # 'Intended Audience :: End Users/Desktop',
    # 'Intended Audience :: Science/Research',
    # 'Intended Audience :: System Administrators',
    # 'License :: Other/Proprietary License',
    # 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    # 'License :: OSI Approved :: MIT License',
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    # 'Programming Language :: Python :: Implementation :: PyPy',
]


########################################################################
# Dependency Specification
########################################################################

PYTHON_REQUIRES = ">=3.6"
PACKAGE_DEPENDENCIES: t.Tuple[str, ...] = ()
SETUP_DEPENDENCIES: t.Tuple[str, ...] = ()
TEST_DEPENDENCIES: t.Tuple[str, ...] = ()
EXTRAS_DEPENDENCIES: t.Dict[str, t.Sequence[str]] = {
    "dev": (
        TEST_DEPENDENCIES
        + (
            "black",
            "coverage",
            "flake8",
            "ipdb",
            "ipython",
            "mypy",
            "pydocstyle",
            "pylint",
            "pytest",
            "pytest-cov",
            "tox",
            "twine",
            "typeguard",
            "wheel",
        )
    )
}


########################################################################
# Package Extras
########################################################################

ENTRY_POINTS: t.Union[str, t.Dict[str, t.Union[str, t.Sequence[str]]]] = {}
PACKAGE_DATA: t.Dict[str, t.Sequence[str]] = {"safetywrap": ["py.typed"]}


########################################################################
# Setup Logic
########################################################################

PACKAGE_DIR = realpath(dirname(__file__))


REQ_FILE = join(PACKAGE_DIR, "requirements_unfrozen.txt")
if exists(REQ_FILE):
    with open(join(PACKAGE_DIR, "requirements.txt")) as reqfile:
        for ln in (l.strip() for l in reqfile):
            if ln and not ln.startswith("#"):
                PACKAGE_DEPENDENCIES += (ln,)


__version__ = "0.0.0"

with open(join(cwd, "src/{}/__init__.py".format(NAME))) as init_file:
    for line in init_file:
        # This will set __version__ and __version_info__ variables locally
        if line.startswith("__version"):
            exec(line)

setup(
    author_email=EMAIL,
    author=AUTHOR,
    classifiers=CLASSIFIERS,
    description=SHORT_DESC,
    entry_points=ENTRY_POINTS,
    extras_require=EXTRAS_DEPENDENCIES,
    keywords=KEYWORDS,
    long_description_content_type=LONG_DESC_CONTENT_TYPE,
    long_description=LONG_DESC,
    name=NAME,
    package_data=PACKAGE_DATA,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=PYTHON_REQUIRES,
    setup_requires=SETUP_DEPENDENCIES,
    tests_require=TEST_DEPENDENCIES,
    url=URL,
    version=__version__,
)
